{{
Copyright (c) 2008 Michael Green.  See end of file for terms of use.
}}

OBJ
   dsp  : "vga_Driver"                 ' VGA Text Driver
   key  : "comboKeyboard"              ' Keyboard Driver
   sto  : "massStorage"                ' EEPROM, SRAM, and Winbond Flash

CON
   version   = 1                       ' Major version
   release   = 0                       ' Minor release
   testLevel = 0                       ' Test change level

   _clkmode  = xtal1 + pll16x
   _xinfreq  = 5_000_000

   sram0CS   =  0                      ' SRAM 0 CS pin number
   sram1CS   =  1                      ' SRAM 1 CS pin number
   winPinDO  =  4                      ' Winbond DO pin number
   winPinClk =  5                      ' Winbond Clk pin number
   winPinDI  =  6                      ' Winbond DI pin number
   winPinCS  =  7                      ' Winbond CS pin number
   tvPins    = 12                      ' TV pin numbers
   vgaPins   = 16                      ' VGA pin numbers
   keyPins   = 26                      ' Normal PS/2 keyboard

   bspKey    = $C8                     ' PS/2 keyboard backspace key
   breakKey  = $CC                     ' PS/2 keyboard application key

   maxName   = 15                      ' Maximum length of an identifier
   maxDict   = $7FFF                   ' Dictionary must be less than this
   lineLen   = dsp#cols                ' Maximum input line length
   endFile   = $FF                     ' Erased byte - end of file mark
   quote     = 34                      ' Double quote
   tabCols   =  8                      ' Number of columns per tab stop
   filesLine = dsp#cols / 14           ' Number of file names on a line
   msgArea   = dsp#cols - 30           ' Start of error message area
   cmdLine   = dsp#rows - 1            ' Command line (and error message)
   cmdArea   =  2                      ' Start of command entry area

   txtFile   =  "t" | "x"<<8 | "t"<<16 ' Extension for text files
   binFile   =  "b" | "i"<<8 | "n"<<16 ' Extension for binary files
   extMask   =  $FF | $FF<<8 | $FF<<16 ' Empty extension

   #0                                  ' Token codes produced by lexical scanner
   tokenIdent                          ' Any identifier (1st 15 chars unique)
   tokenString                         ' Any string value (length up to maxLine)
   tokenNumber                         ' Any integer value (1 to 4 bytes)

VAR
   long ioControl[2], fcbIn[4], fcbOut[4], lineNo, tokenPtr
   byte scanPtr, scanLine[lineLen+1]

PUB main | err, s, state
'' Clear the program space and variables, then read a line and interpret it.
   dsp.start(vgaPins)                  ' Start display
   sto.winSetVideo(dsp.videoCog(0),dsp.videoCog(1))
   key.start(keyPins)                  ' Start keyboard driver
   sto.start(@ioControl)               ' Start I2C/SPI driver
   waitcnt(clkfreq + cnt)
   dsp.str(string("Ouroboros Scanner"))
   if version > 0 or release > 0 or testLevel > 0
     dsp.str(string(" Version "))
     dsp.dec(version)
     dsp.out(".")
     if release < 100
       dsp.out("0")
     if release < 10
       dsp.out("0")
     dsp.dec(release)
     if testLevel > 0
       dsp.out("a"+testLevel-1)
   dsp.str(string(dsp#Cr,dsp#Lf))
   dsp.str(string("Free space: "))
   dsp.dec((dsp#sync - @s - 4)>>10)
   dsp.str(string("K",dsp#Cr,dsp#Lf))
   dsp.str(string("Mounting flash ... "))
   dsp.dec(sto.winStart(winPinCS,winPinClk,winPinDI,winPinDO,sram0CS,sram1CS)*4)
   dsp.str(string("K free out of "))
   dsp.dec(sto.flashSize>>20)
   dsp.str(string("MB",dsp#Cr,dsp#Lf,dsp#Esc,"[3z"))
   waitcnt(clkfreq + cnt)
   if not key.present
      dsp.str(string("No keyboard present",dsp#Cr))
      abort
   key.clearkeys
   dsp.clearRegion(0,0,dsp#cols,cmdLine) ' Clear display area
   repeat
      dsp.setCursor(0,cmdLine)         ' Return true if yes/YES/y/Y
      dsp.str(string(">>"))            ' Return false otherwise
      state~
      repeat
         state := windowEditor(cmdArea,cmdLine,msgArea-cmdArea-1,1,state)
      until state & $FF == $0D
      dsp.clearRegion(msgArea,cmdLine,30,1) ' Clear message area & get text
      dsp.getLine(@scanLine,cmdArea,cmdLine,msgArea-cmdArea-1)
      ifnot scanLine[0]
         listFiles                     ' No file name?  List directory
         next
      if sto.initFile(@fcbIn,@scanLine)
         if fcbIn[2] & extMask == binFile ' Extension given was ".bin"
            ifnot sto.openFile(@fcbIn) ' Try to open the file name
               dsp.setLine(msgArea,cmdLine,string("Can't open executable"))
               next
            sto.bootFile(@fcbIn)       ' Execute the file with Spin
         if fcbIn[2] & extMask == extMask ' No extension was provided
            if sto.openFile(@fcbIn)    ' Try to open the file name
               sto.bootFile(@fcbIn)    ' Assume it's an executable
            fcbIn[2] := binFile        ' Can't open, try XXX.bin
            if sto.openFile(@fcbIn)
               sto.bootFile(@fcbIn)    ' Execute the file if present
            fcbIn[2] := txtFile        ' Try XXX.txt also
         if sto.openFile(@fcbIn)
            quit
         else
            dsp.setLine(msgArea,cmdLine,string("Can't open file"))
            next
      else
         dsp.setLine(msgArea,cmdLine,string("Invalid filename"))
   sto.initFile(@fcbOut,string("workfile.tmp"))
   ifnot sto.createFile(@fcbOut)       ' Try to create a temporary file
      ifnot sto.eraseFile(@fcbOut)     ' If already exists, erase it
         dsp.setLine(msgArea,cmdLine,string("Internal error (erase)"))
      else
         ifnot sto.createFile(@fcbOut) '  and try to create it again
            dsp.setLine(msgArea,cmdLine,string("Can't create file"))
         else
            initCompiler               ' Initialize the compiler
            repeat until s := fillScanLine
               if s := \processLine
                  quit                 ' Stop on error or end of file
            dsp.setLine(msgArea,cmdLine,||s)
   if s == 0                           ' If no errors, go on to phase 2
      if sto.initFile(@fcbIn,string("Ourobor2.bin"))
         if sto.openFile(@fcbIn)
            sto.bootFile(@fcbIn)
   if sto.initFile(@fcbIn,string("femtobas.bin"))
      if sto.openFile(@fcbIn)          ' Use FemtoBasic if errors occurred
         sto.bootFile(@fcbIn)
   reboot

PRI initCompiler                       ' Initialize phase 1 of the compiler
   lineNo~
   tokenPtr~                           ' Address of next token in SRAM

PRI processLine | t, s, f0,f1,f2,f3    ' Scan one line of the source program
   lineNo++
   if scanPtr == 0 and scanLine[0] == endFile
      return string("End of program")  ' End of file found
   repeat
      skipSpaces
      case scanLine[scanPtr]
         "A".."Z","_","a".."z":
            scanName(@f0)
            t := tokenIdent
            s := lookupToken(@f0,strsize(@f0))
         quote:
            if (result := scanString(@f0)) < 0
               quit
            t := tokenString
            s := lookupToken(f0,strsize(f0))
         "$","%","0".."9":
            if (result := scanNumber(@f0)) < 0
               quit
            t := tokenNumber
            s := lookupToken(@f0,4)
      sto.writeSRAM(@fcbOut,lineNo,-2)
      sto.writeSRAM(@fcbOut,scanPtr,-1)
      sto.writeSRAM(@fcbOut,t,-1)
      sto.writeSRAM(@fcbOut,s,-2)

PRI lookupToken(p,s) | t, i            ' Look up token in SRAM dictionary
   repeat while result < tokenPtr
      if (t := sto.readSRAM(result,0,-1)) == s
         repeat i from 0 to s-1
            if byte[p][i] <> sto.readSRAM(result+i+1,0,-1)
               quit
         if i == s                     ' Complete match.  Return offset
            return
      result += t + 1                  ' On to next entry in dictionary
      next
   sto.writeSRAM(result,s,-1)          ' Not found, make new entry
   sto.writeSRAM(result+1,p,s)
   tokenPtr += s + 1                   ' Advance pointer to next available
   if tokenPtr > maxDict
      abort -string("Dictionary full")

PRI listFiles | a, width, c            ' Display flash directory
   dsp.clearRegion(0,0,dsp#cols,cmdLine) ' Clear display area
   dsp.setCursor(0,0)
   dsp.str(string("Current Flash Files",dsp#Cr,dsp#Lf,dsp#Lf))
   sto.firstFile(@fcbIn)
   c := filesLine                      ' 5 file names per line = 70 cols
   repeat while sto.nextFile(@fcbIn)
      width := 2                       ' 2 spaces between file names
      repeat a from 0 to 7
         if fcbIn.byte[a] == $FF       ' Trailing pad for $FF
            width++
         else
            dsp.out(fcbIn.byte[a])
      if fcbIn[2] & $00FFFFFF == $00FFFFFF
         width++                       ' Trailing pad for no "."
      else
         dsp.out(".")
      repeat a from 8 to 10
         if fcbIn.byte[a] == $FF       ' Trailing pad for $FF
            width++
         else
            dsp.out(fcbIn.byte[a])
      if --c                           ' Don't pad after last file name
         repeat width
            dsp.out(" ")
      else
         dsp.str(string(dsp#Cr,dsp#Lf))
         c := filesLine                ' Start a new display line
   if c <> filesLine
      dsp.str(string(dsp#Cr,dsp#Lf))   ' Finish any partial line

PRI fillScanLine | c                   ' Fill scanLine with the next line
   scanPtr~
   repeat
      case c := sto.readSRAM(@fcbIn,0,-1)
         endFile:
            if scanPtr > 0             ' End of file must come after CR/LF
               return string("Unexpected end of file")
            bytemove(@scanLine,string(endFile),2)
            return false
         dsp#Cr:
            scanLine[scanPtr]~         ' Carriage return marks end of string
            return false
         dsp#Lf:
            next                       ' Ignore line feeds
         other:
            if scanPtr == lineLen
               return string("Input line too long")
            scanLine[scanPtr++] := c

PRI skipSpaces                         ' Skip leading spaces
   repeat while (result := scanLine[scanPtr]) == " "
      scanPtr++

PRI scanName(p) | c                    ' Scan identifier into buffer at p
   bytefill(p,0,maxName+1)             ' Clear name buffer
   case c := scanLine[scanPtr]         ' First character: letter or underscore
      "A".."Z", "_", "a".."z":
         byte[p++] := c
         repeat                        ' Subsequent characters: digits too
            case c := scanLine[++scanPtr]
               "0".."9", "A".."Z", "_", "a".."z":
                  if result < maxName  ' At most maxName characters
                     result++
                     byte[p++] := c
               other:
                  byte[p]~
                  quit
      other:
         return false

PRI scanString(p) | c, pd              ' Return the index to scanLine past the
   if scanLine[scanPtr] <> quote       '  end of a scanned string.  Return < 0
      return -string("String expected") '  if an error occured.  Convert in
   long[p] := pd := scanPtr            '    place the string to its content
   repeat                              '     with a null terminating byte.
      case c := scanLine[++scanPtr]
         quote: scanPtr++              ' Next quote ends the string
                scanLine[pd]~
                quit
         "\":   case c := scanLine[++scanPtr]
                   "b":   scanLine[pd++] := dsp#Bsp ' \b is a backspace
                   "e":   scanLine[pd++] := dsp#Esc ' \e is an escape char
                   "n":   scanLine[pd++] := dsp#Lf  ' \n is a newline (LF)
                   "r":   scanLine[pd++] := dsp#Cr  ' \r is a return
                   "t":   scanLine[pd++] := dsp#Tab ' \t is a tab
                   0:     scanPtr--
                          return -string("End of line after \")
                   other: scanLine[pd++] := c       ' escape like \\ or \"
         0:     scanPtr--              ' Leave pointer at end of line
                return -string("End of line in string")
         other: scanLine[pd++] := c    ' Copy the character otherwise

PRI scanNumber(p) | c, d, t            ' Scan numeric value into long at p
   case c := scanLine[scanPtr]
      quote:                           ' String as 1 to 4 byte value
         if (d := scanString(@t)) < 0
            return d                   ' Error scanning string
         if strsize(t) > 4
            return -string("String as number > 32 bits")
         long[p]~                      ' Copy value to result
         bytemove(p,@scanLine+t,strsize(t))
      "$":                             ' Hexadecimal value
         c := scanLine[++scanPtr]
         if (t := lookup(c:"0123456789abcdef")) == 0
            return -string("Invalid hex character")
         long[p] := t
         c := scanLine[++scanPtr]      ' Allow embedded underlines
         d := c | (c & $40) >> 1       ' Make lower case
         repeat while (t := lookup(d:"0123456789abcdef")) == 0 or c == "_"
            if c <> "_"
               if long[p] & $F0000000
                  return -string("Hexadecimal value > 32 bits")
               long[p] := long[p] << 4 | (t - 1)
            c := scanLine[++scanPtr]
            d := c | (c & $40) >> 1    ' Make lower case
      "%":
         c := scanLine[++scanPtr]
         if not (c == "0" or c == "1")
            return -string("Invalid binary character")
         long[p] := c - "0"
         c := scanLine[++scanPtr]      ' Allow embedded underlines
         repeat while (c == "0" or c == "1") or c == "_"
            if c <> "_"
               if long[p] & $80000000
                  return -string("Binary value > 32 bits")
               long[p] := long[p] << 1 | (c - "0")
            c := scanLine[++scanPtr]
      "0".."9":
         long[p] := c - "0"
         c := scanLine[++scanPtr]      ' Allow embedded underlines
         repeat while (c => "0" and c =< "9") or c == "_"
            if c <> "_"
               if long[p] > 214748364 or long[p] == 214748364 and c > "7"
                  return -string("Decimal value > 31 bits")
               long[p] := long[p] * 10 + c - "0"
            c := scanLine[++scanPtr]
      other:
        return -string("Invalid literal value")

PUB dumpMemory(pin,addr,size) | i, c, p, first, buf0, buf1, buf2
'' This routine dumps a portion of the RAM/ROM to the display (pin == -1).
'' If pin is not -1, it is an EEPROM address in the form required by the
'' I2C routines in i2cSpiInit.  The specified address is or'd with this.
'' The format is 8 bytes wide with hexadecimal and ASCII.  If the address
'' is greater than $7FFFFF, then it's treated as a flash memory address
'' that starts at $800000 (translated to $000000 to $7FFFFF).  If the
'' address is greater than $FFFF, but less than $800000, it's treated
'' as an SRAM address from $0000 to $FFFF.
   first := true
   p := addr & $FFFFF8
   repeat while p < (addr + size)
      if first
         dsp.hex(addr,6)
         first := false
      else
         dsp.hex(p,6)
      dsp.out(":")
      repeat i from 0 to 7
         byte[@buf0][i] := " "
         if p => addr and p < (addr + size)
            c := 0
            if pin <> -1
               if sto.readEEPROM(pin|p,@c,1)
                  abort string("EEPROM read")
            else
               if p > $7FFFFF
                  sto.readData(p & $7FFFFF,@c,1)
               else
                  if p > $FFFF
                     sto.readSRAM(p & $FFFF,@c,1)
                  else
                     c := byte[p]
            dsp.hex(c,2)
            if c => " " and c =< "~"
               byte[@buf0][i] := c
         else
            dsp.out(" ")
            dsp.out(" ")
         if i < 7
            dsp.out(" ")
         p++
      buf2 := 0
      dsp.out("|")
      dsp.str(@buf0)
      dsp.out("|")
      dsp.str(string(dsp#Cr,dsp#Lf))

PUB windowEditor(winX, winY, winC, winR, state) | c, t, relCol, relRow

' This method provides an editable window on the display text screen.
' It uses the standard PS/2 keyboard control keys including backspace,
' cursor positioning, delete and insert (for an insert mode).
' The edited text must be retrieved from the window by the caller.
' Unknown or invalid control keys terminate editing and return the keycode.
' The state of the window and the screen is such that editing can be
' resumed after the terminating keycode is processed by the caller.
' The return value contains: insert << 31 | y << 16 | x << 8 | keycode
' and is normally used on a subsequent call to maintain the state.
' A value of zero is used to initialize the editor to the upper left
' corner of the window.  No checking is done for valid window bounds.

   if state == 0
      dsp.clearRegion(winX,winY,winC,winR)
   dsp.setCursor(((state>>8)&$FF)+winX,((state>>16)&$FF)+winY)
   if state ~>= 31
      dsp.cursorMode(dsp#fastCurs,true)
   else
      dsp.cursorMode(dsp#slowCurs,true)
   repeat
      relCol := dsp.getCol - winX
      relRow := dsp.getRow - winY
      case c := key.getkey
         $09:                          ' Tab
            repeat
               if state                ' Insert spaces to next tab stop
                  dsp.insertSpaces(dsp.getCol,dsp.getRow,1)
               dsp.out(" ")
            while (dsp.getCol - winX) // tabCols
         $0D:                          ' Return
            if state                   ' Insert mode
               return state<<31 | relRow<<16 | relCol<<8 | c
            else                       ' Normal mode
               dsp.clearRegion(dsp.getCol,dsp.getRow,winC-relCol,1)
               dsp.setCol(winX)        ' Go to beginning of next line
               relCol := dsp.getCol - winX
               relRow := dsp.getRow - winY
               if relRow == winR-1     ' Exit if at bottom of window
                  return state<<31 | relRow<<16 | relCol<<8 | c
               dsp.out(dsp#Lf)
         $20..$7E:                     ' Printable characters
            if relCol < winC-1
               if state
                  dsp.insertSpaces(winX+relCol,winY+relRow,1)
               dsp.out(c)
         $C0:                          ' Left arrow
            if relCol > 0
               dsp.str(string(dsp#Esc,"[D"))
         $C1:                          ' Right arrow
            if relCol < winC-1
               dsp.str(string(dsp#Esc,"[C"))
         $C2:                          ' Up arrow
            if relRow == 0             ' Exit if already at top of window
               return state<<31 | relRow<<16 | relCol<<8 | c
            if winR == 1               ' Special case for a single line window
               return state<<31 | relRow<<16 | relCol<<8 | c
            dsp.str(string(dsp#Esc,"[A"))
         $C3:                          ' Down arrow
            if relRow == winR-1        ' Exit if already at bottom of window
               return state<<31 | relRow<<16 | relCol<<8 | c
            if winR == 1               ' Special case for a single line window
               return state<<31 | relRow<<16 | relCol<<8 | c
            dsp.str(string(dsp#Esc,"[B"))
         $C4:                          ' Home
            if winR == 1               ' Special case for a single line window
               return state<<31 | relRow<<16 | relCol<<8 | c
            dsp.setCursor(0,0)
         $C5:                          ' End
            if winR == 1               ' Special case for a single line window
               return state<<31 | relRow<<16 | relCol<<8 | c
            dsp.setCursor(0,winY+winR-1)
         $C8:                          ' Backspace
            if relCol > 0
               dsp.deleteText(winX+relCol-1,winY+relRow,1)
               dsp.setCol(winX+relCol-1)
            else                       ' Already at left end of line
               return state<<31 | relRow<<16 | relCol<<8 | c
         $C9, $7E:                     ' Delete
            if relCol == dsp#cols-1
               return state<<31 | relRow<<16 | relCol<< 8 | c
            else                       ' Already at right end of line
               dsp.deleteText(winX+relCol,winY+relRow,1)
         $CA:                          ' Toggle Insert mode
            if state := not state
               dsp.cursorMode(dsp#fastCurs,true)
            else
               dsp.cursorMode(dsp#slowCurs,true)
         other:                        ' All others, exit
            return state<<31 | relRow<<16 | relCol<<8 | c

{{
                            TERMS OF USE: MIT License

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
}}

